import * as THREE from '../build/three.module.js';

import Stats from './jsm/libs/stats.module.js';
import { GUI } from './jsm/libs/dat.gui.module.js';

import { OrbitControls } from './jsm/controls/OrbitControls.js';
import { OBJLoader } from './jsm/loaders/OBJLoader.js';
import { EffectComposer } from './jsm/postprocessing/EffectComposer.js';
import { RenderPass } from './jsm/postprocessing/RenderPass.js';
import { ShaderPass } from './jsm/postprocessing/ShaderPass.js';
import { OutlinePass } from './jsm/postprocessing/OutlinePass.js';
import { FXAAShader } from './jsm/shaders/FXAAShader.js';

class App {
    constructor() {
        this._initialize();
    }

    _initialize() {
        this.domWebGL = document.createElement('div');
        document.body.appendChild(this.domWebGL);

        let renderer = new THREE.WebGLRenderer({ antialias: true });
        renderer.shadowMap.enabled = true;
        renderer.domElement.style.touchAction = "none";
        renderer.domElement.addEventListener("pointermove", this._onPointerMove.bind(this));
        this.domWebGL.appendChild(renderer.domElement);

        this.renderer = renderer;

        this.scene = new THREE.Scene();

        this._stats = new Stats();
        this.domWebGL.appendChild(this._stats.dom);

        this._raycaster = new THREE.Raycaster();
        this._mouse = new THREE.Vector2();

        this._setupCamera();
        this._setupGUI();
        this._setupControls();
        this._setupLights();
        this._setupModel();
        this._setupPostprocess();

        window.onresize = this.resize.bind(this);


        this.resize();
    }

    _onPointerMove(event) {
        if(event.isPrimary === false) return;

        this._mouse.x = (event.clientX / window.innerWidth) * 2 - 1;
        this._mouse.y = -(event.clientY / window.innerHeight) * 2 + 1;
        
        this._checkIntersection();
    }

    _checkIntersection() {
        this._raycaster.setFromCamera(this._mouse, this._camera);

        const intersects = this._raycaster.intersectObject(this.scene, true);
        if(intersects.length > 0) {
            const selectedObject = intersects[0].object;
            this._outlinePass.selectedObjects = [ selectedObject ];
        } else {
            //this._outlinePass.selectedObjects = [];
        }
    }

    _setupCamera() {
        const camera = new THREE.PerspectiveCamera(45, window.innerWidth / window.innerHeight, 0.1, 100);
        camera.position.set(0, 0, 8);
        this._camera = camera;
    }

    _setupControls() {
        const controls = new OrbitControls(this._camera, this.renderer.domElement);
        controls.minDistance = 5;
        controls.maxDistance = 20;
        controls.enablePan = false;
        controls.enableDamping = true;
        controls.dampingFactor = 0.05;

        this._controls = controls;
    }

    _setupLights() {
        this.scene.add(new THREE.AmbientLight(0xaaaaaa, 0.2));

        const light = new THREE.DirectionalLight(0xddffdd, 0.6);
        light.position.set(1, 1, 1);
        light.castShadow = true;
        light.shadow.mapSize.width = 1024;
        light.shadow.mapSize.height = 1024;

        const d = 10;
        light.shadow.camera.left = -d;
        light.shadow.camera.right = d;
        light.shadow.camera.top = d;
        light.shadow.camera.bottom = -d;
        light.shadow.camera.far = 1000;

        this.scene.add(light);



    }

    _setupModel() {
        const manager = new THREE.LoadingManager();
        manager.onProgress = (item, loaded, total) => {
            console.log(item, loaded, total);
        };

        const obj3d = new THREE.Object3D();

        const loader = new OBJLoader(manager);
        loader.load("models/obj/tree.obj", (object) => {
            let scale = 1.0;
            object.traverse((child) => {
                if(child instanceof THREE.Mesh) {
                    child.geometry.center();
                    child.geometry.computeBoundingSphere();
                    scale = 0.2 * child.geometry.boundingSphere.radius;

                    const phongMaterial = new THREE.MeshPhongMaterial( { color: 0xffffff, specular: 0x111111, shininess: 5 });
                    child.material = phongMaterial;
                    child.receiveShadow = true;
                    child.castShadow = true;
                }
            });

            object.position.y = 1;
            object.scale.divideScalar(scale);
            obj3d.add(object);
        });

        const group = new THREE.Group();
        this.scene.add(group);
        group.add(obj3d);

        const geometry = new THREE.SphereGeometry(3, 48, 24);
        for(let i=0; i<20; i++) {
            const material = new THREE.MeshLambertMaterial();
            material.color.setHSL(Math.random(), 1.0, 0.3);

            const mesh = new THREE.Mesh(geometry, material);
            mesh.position.x = Math.random() * 4 - 2;
            mesh.position.y = Math.random() * 4 - 2;
            mesh.position.z = Math.random() * 4 - 2;
            mesh.receiveShadow = true;
            mesh.castShadow = true;
            mesh.scale.multiplyScalar(Math.random() * 0.3 + 0.1);
            group.add(mesh);
        }

        const floorMaterial = new THREE.MeshLambertMaterial( { side: THREE.DoubleSide });
        const floorGeometry = new THREE.PlaneGeometry(12, 12);
        const floorMesh = new THREE.Mesh(floorGeometry, floorMaterial);
        floorMesh.rotation.x -= Math.PI * 0.5;
        floorMesh.position.y -= 1.5;
        group.add(floorMesh);
        floorMesh.receiveShadow = true;

        const torusGeometry = new THREE.TorusGeometry(1, 0.3, 16, 100);
        const torusMaterial = new THREE.MeshPhongMaterial({ color: 0xffaaff });
        const torus = new THREE.Mesh(torusGeometry, torusMaterial);
        torus.position.z = -4;
        group.add(torus);
        torus.receiveShadow = true;
        torus.castShadow = true;

        this._group = group;
    }

    _setupPostprocess() {
        const composer = new EffectComposer(this.renderer);
        
        const renderPass = new RenderPass(this.scene, this._camera);
        composer.addPass(renderPass);

        const outlinePass = new OutlinePass(new THREE.Vector2(window.innerWidth, window.innerHeight), this.scene, this._camera);
        composer.addPass(outlinePass);

        const textureLoader = new THREE.TextureLoader();
        textureLoader.load("textures/tri_pattern.jpg", (texture) => {
            outlinePass.patternTexture = texture;
            texture.wrapS = THREE.RepeatWrapping;
            texture.wrapT = THREE.RepeatWrapping;
        });

        const effectFXAA = new ShaderPass(FXAAShader);
        effectFXAA.uniforms["resolution"].value.set(1/window.innerWidth, 1/window.innerHeight);
        composer.addPass(effectFXAA);

        this._outlinePass = outlinePass;
        this._composer = composer;
        this._effectFXAA = effectFXAA;
    }

    update(time) {
        if(this._params.rotate) {
            this._group.rotation.y = time*0.0001;
        }

        this._controls.update();
    }

    render(time) {
        requestAnimationFrame(this.render.bind(this));

        this._stats.begin();

        this._composer.render();
        this.update(time);

        this._stats.end();
    }

    resize() {
        this._camera.aspect = window.innerWidth / window.innerHeight;
        this._camera.updateProjectionMatrix();

        this.renderer.setPixelRatio(window.devicePixelRatio);
        this.renderer.setSize(window.innerWidth, window.innerHeight);
        this._composer.setSize(window.innerWidth*window.devicePixelRatio, window.innerHeight*window.devicePixelRatio);
        this._effectFXAA.uniforms["resolution"].value.set(1/(window.innerWidth*window.devicePixelRatio), 1/(window.innerHeight*window.devicePixelRatio));
    }

    _setupGUI() {
        this._params = {
            edgeStrength: 3.0,
            edgeGlow: 0.0,
            edgeThickness: 1.0,
            pulsePeriod: 0,
            rotate: false,
            usePatternTexture: false,
            visibleEdgeColor: "#ffffff",
            hiddenEdgeColor: "#190a05"
        };

        const gui = new GUI({ width:300 });

        gui.add(this._params, "edgeStrength", 0.01, 10).onChange((value) => {
            this._outlinePass.edgeStrength = Number(value);
        });

        gui.add(this._params, "edgeGlow", 0.0, 1).onChange((value) => {
            this._outlinePass.edgeGlow = Number(value);
        });

        gui.add(this._params, "edgeThickness", 1, 4).onChange((value) => {
            this._outlinePass.edgeThickness = Number(value);
        });

        gui.add(this._params, "pulsePeriod", 0.0, 5).onChange((value) => {
            this._outlinePass.pulsePeriod = Number(value);
        });

        gui.add(this._params, "rotate");

        gui.add(this._params, "usePatternTexture").onChange((value) => {
            this._outlinePass.usePatternTexture = value;
        })

        gui.addColor(this._params, "visibleEdgeColor").onChange((value) => {
            this._outlinePass.visibleEdgeColor.set(value);
        });

        gui.addColor(this._params, "hiddenEdgeColor").onChange((value) => {
            this._outlinePass.hiddenEdgeColor.set(value);
        });
    }
}

window.onload = function () {
    (new App()).render();
}